/*
* parse diff files
* 
* Author: Zelea
* Date: 12 July 2011
*/

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>

#define BUF_SIZE 0x400
#define MAX 1000

struct patch_e
{
  unsigned      addr;
  unsigned char old;
  unsigned char new;
} __attribute__((__packed__));

struct patch_e *ptable;

static void
read_diff( char *fname )
{
  FILE         *dif, *fh;
  char         *p, *hfile = "patch.h";
  char          ln[BUF_SIZE];
  unsigned char c;
  int		before, after;
  unsigned char *b;
  struct patch_e *P;
  int		start[MAX/2];
  int           i, l, pfile;

  ptable = ( struct patch_e * ) malloc( MAX * sizeof( struct patch_e ) );
  if ( ptable == NULL )
  {
    fprintf( stderr, "Cannot allocate memory for patches\n" );
    return;
  }
  if ( ( dif = fopen( fname, "r" ) ) == NULL )
  {
    perror( fname );
    return;
  }
  if ( ( fh = fopen( hfile, "w" ) ) == NULL )
  {
    perror( hfile );
    return;
  }
  printf( "\nCreating \"%s\" file from diff file \"%s\"\n\n", hfile, fname );
  fprintf( fh, "char *AUTHOR = \"\";\nchar *TITLE = \"\";\nchar *VERIFIED = \"\";\n\n" );
  fprintf( fh, "struct patch_e\n{\nunsigned addr;\nunsigned char old;\nunsigned char new;\n} __attribute__((__packed__));\n\n" );
  fprintf( fh, "char         *execs[] = { " );
  i = pfile = 0;
  P = ptable;
  while ( fgets( ln, BUF_SIZE, dif ) != NULL )
  {
    l = strlen( ln );
    do
    {
      c = ln[l - 1];
      if ( c == '\n' || c == '\r' || c == ' ' || c == '\t' )
	ln[--l] = c = 0;
    }
    while ( l && c == 0 );
    if ( !l )
      continue;
    if( ( p = strstr( ln, " and " ) ) != NULL )
    {
      *p = 0;
      if( ( p = strrchr( ln, ' ' ) ) == NULL )
	continue;
      p++;
      fprintf( fh, "\"%s\", ", p );
      start[pfile++] = i;
      continue;
    }
    // store the patches 
    if ( sscanf( ln, "%x: %02x %02x", &P->addr, &before, &after ) == 3 )
    {
      P->old = before;
      P->new = after;
      i++;
      if ( i == MAX )
      {
	fprintf( stderr, "To many patches for file \"%s\"\n", fname );
	return;
      }
      P++;
    }
  }
  fclose( dif );
  start[pfile] = i;
  printf( "%d bytes to patch in %d files\n", i, pfile );
  fprintf( fh, "NULL };\n\n" );
  fprintf( fh, "int           patch_off[] = { " );
  for( l = 0; l < pfile; l++ )
    fprintf( fh, "%d,%d,", start[l], start[l+1] - start[l] );
  fprintf( fh, "};\n\n" );
  fprintf( fh, "unsigned char patch[] = { " );
  b = ptable;
  for( l = 0; l < i * sizeof( struct patch_e ); l++ )
  {
    if( *b <= 9 )
      fprintf( fh, "%d,", *b++ );
    else
      fprintf( fh, "0x%x,", *b++ );
  }
  fprintf( fh, "};\n\n" );
  fclose( fh );
  if ( ptable )
  {
    free( ptable );
    ptable = NULL;
  }
}

static void
usage( char *progname )
{
  fprintf( stderr, "\nUsage:\t%s diff_file(s)\n\n", progname );
  exit( -1 );
}

int
main( int argc, char *argv[] )
{
  if ( argc < 2 )
    usage( argv[0] );
  read_diff( argv[1] );
  exit( 0 );
}
