#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <dirent.h>
#include <string.h>
#include <utime.h>

#include "patch.h"
#include "crack.h"

#define PATHSZ 1024

char          cpath[PATHSZ];
int           total_patches, topdirlen, dry_run, reverse_patch;

unsigned char bbuf[PATHSZ];

int
patch_diff( char *binpath, int name_off, int offset )
{
  FILE         *f;
  char         *ftree;
  int           matches;
  long          boff;
  int          *para, i, o, start, numpatch, l;
  unsigned char old, new;
  struct utimbuf tmst;
  struct stat   flst;
  struct patch_e *P;

  // get the file status
  if ( stat( binpath, &flst ) == -1 )
    return -1;
  if ( !S_ISREG( flst.st_mode ) )	// check only regular files
    return -1;
  ftree = binpath + name_off;
  if ( !dry_run && !( flst.st_mode & S_IWUSR ) )	// set the write permission 
  {
    if ( !chmod( binpath, ( flst.st_mode & ( ( 1 << 12 ) - 1 ) ) | S_IWUSR ) )
      if ( ftree )
	printf( "Adding write permission to file \"%s\"\n", ftree );
  }
  // open the file in read or read-write mode
  f = fopen( binpath, dry_run ? "rb" : "r+b" );
  if ( f == NULL )
  {
    printf( "File \"%s\" not found (or no write permissions)\n\n", ftree );
    return -2;
  }
  printf( "--> Patching file: %s\n", ftree );
  para = patch_off + 2 * offset;
  start = *para++;
  numpatch = *para;
  P = ( struct patch_e * ) patch;
  P += start;
  boff = -PATHSZ;
  for ( l = i = 0; i < numpatch; P++, i++ )
  {
    if ( P->addr - boff >= PATHSZ )
    {
      if ( !dry_run && boff >= 0 && l )
      {
	fseek( f, boff, SEEK_SET );
	fwrite( bbuf, 1, l, f );
      }
      boff = P->addr;
      fseek( f, boff, SEEK_SET );
      l = fread( bbuf, 1, PATHSZ, f );
    }
    o = P->addr - boff;
    if ( reverse_patch )
    {
      old = P->new;
      new = P->old;
    }
    else
    {
      old = P->old;
      new = P->new;
    }
    if ( bbuf[o] != old )
      break;
    bbuf[o] = new;
    if ( verbose )
      printf( "%08x: %02x %02x\n", P->addr, old, new );
  }
  if ( !dry_run && boff >= 0 && l )
  {
    fseek( f, boff, SEEK_SET );
    fwrite( bbuf, 1, l, f );
  }
  if ( i == numpatch )
    matches = numpatch;
  else
    matches = 0;
  fclose( f );

  if ( matches )
  {
    // try to restore the timestamps
    if ( !dry_run )
    {
      tmst.actime = flst.st_atime;
      tmst.modtime = flst.st_mtime;
      utime( binpath, &tmst );
    }
    printf( "%d bytes patched\n", matches );
  }
  else if ( verbose > 1 )
    printf( "no match found\n" );
  return matches;
}

#define CASEFOLD 0x10
int           fnmatch( const char *pattern, const char *string, int flags );

void
scan_dir( char *dirname, char **match, int *function(  ), void *result )
{
  DIR          *dir;
  struct dirent *fd;
  struct stat   fstat;
  int           i, r;

  if ( chdir( dirname ) == -1 )
    return;
  dir = opendir( dirname );
  while ( ( fd = readdir( dir ) ) != NULL )
  {
    if ( fd->d_name[0] == '.' )
      continue;
    if ( stat( fd->d_name, &fstat ) != 0 )
      continue;
    if ( S_ISDIR( fstat.st_mode ) )
    {
      if ( chdir( fd->d_name ) != -1 )
      {
	scan_dir( ".", match, function, result );
	chdir( ".." );
	continue;
      }
    }
    for ( i = 0; match[i] != NULL; i++ )
      if ( fnmatch( match[i], fd->d_name, CASEFOLD ) == 0 )
	break;
    if ( match[i] == NULL )
      continue;
    if ( ( r = ( int ) function( fd->d_name, i ) ) > 0 )
      total_patches += r;
  }
  closedir( dir );
}

int
crack( char *binname, int offset )
{
  int           r;

  if ( getcwd( cpath, PATHSZ ) == NULL )
    return -1;
  r = strlen( cpath );
  cpath[r++] = SLASH;
  strcpy( cpath + r, binname );
  return patch_diff( cpath, r, offset );
}

int
flow( char *topdir )
{
  chdir( topdir );
  if ( getcwd( cpath, PATHSZ ) == NULL )
    return -1;
  topdirlen = strlen( cpath );
  printf( "Scanning path: %s\n\n", cpath );
  scan_dir( cpath, execs, ( void * ) crack, NULL );
  printf( "\n" );
  return 1;
}
